"""Provides short-hand functions for 'CreateFile', 'SetFile' IOC classes with runtime config"""

import logging
from typing import TYPE_CHECKING

import nawah.data as Data
from nawah.config import Config

from ._classes import CreateFile, SetFile

if TYPE_CHECKING:
    from nawah.types import NawahDoc, NawahSession, Results

logger = logging.getLogger("nawah")


async def create_file(
    *,
    module_name: str,
    session: "NawahSession",
    doc: "NawahDoc",
) -> "Results":
    """Creates file doc for a module"""

    # This is short-hand wrapper around CreateFile class which injects Config.modules, Data.create

    create_file_implementation = CreateFile(
        modules=Config.modules,
        data_create_callable=Data.create,
        set_file_callable=set_file,
    )

    return await create_file_implementation(
        module_name=module_name, session=session, doc=doc
    )


async def set_file(
    *,
    module_name: str,
    session: "NawahSession",
    doc: "NawahDoc",
    raise_no_success: bool,
) -> "Results":
    """Sets doc value of file created using 'create_file'"""

    # This is short-hand wrapper around SetFile class which injects Config.modules, Data.update

    update_file_implementation = SetFile(
        modules=Config.modules, data_update_callable=Data.update
    )

    return await update_file_implementation(
        module_name=module_name,
        session=session,
        doc=doc,
        raise_no_success=raise_no_success,
    )



"""Provides types used in Nawah"""

from ._data import DataCreateCallable, DataUpdateCallable
from ._types import (AppPackage, NawahConn,
                     NawahDoc, NawahEvents, NawahQueryOperAnd,
                     NawahQueryOperOr, NawahQuerySpecial,
                     NawahQuerySpecialGroup, NawahQueryStep, NawahSession,
                     Results, ResultsArgs)

__all__ = [
    "DataCreateCallable",
    "DataUpdateCallable",
    "AppPackage",
    "NawahConn",
    "NawahDoc",
    "NawahEvents",
    "NawahQueryOperAnd",
    "NawahQueryOperOr",
    "NawahQuerySpecial",
    "NawahQuerySpecialGroup",
    "NawahQueryStep",
    "NawahSession",
    "Results",
    "ResultsArgs",
]



"""Provides loops to keep asyncio running, running minutely checks"""

import asyncio
import logging
import sys
import time
from typing import TYPE_CHECKING

import aiohttp.web

import nawah.data as Data
from nawah.config import Config

from ._common_handlers import (_not_allowed_handler, _not_found_handler,
                               _root_handler)
from ._http_handler import _http_handler
from ._utils import _populate_routes

if TYPE_CHECKING:
    from aiohttp.web import Application

logger = logging.getLogger("nawah")


def _create_error_middleware(overrides):
    @aiohttp.web.middleware
    async def error_middleware(request, handler):
        try:
            response = await handler(request)
            override = overrides.get(response.status)
            if override:
                return await override(request)
            return response
        except aiohttp.web.HTTPException as ex:
            override = overrides.get(ex.status)
            if override:
                return await override(request)
            raise

    return error_middleware


async def _web_loop(app: "Application", /):
    # AsnciIOMotor must be created within scope of asyncio.run to be created in same event loop
    # as AIOHTTP loop
    # [TODO] Figure out another callable to call Data.create_conn, after dropping config_app Utility
    Config.sys.conn = Data.create_conn()
    runner = aiohttp.web.AppRunner(app)
    await runner.setup()
    site = aiohttp.web.TCPSite(runner, "0.0.0.0", Config.port)
    await site.start()
    logger.info("Serving on 0.0.0.0:%s", Config.port)
    while True:
        await asyncio.sleep(60)


def _run_app():
    get_routes, post_routes = _populate_routes()

    logger.debug(
        "Loaded modules: %s",
        {module_name: module.attrs for module_name, module in Config.modules.items()},
    )
    logger.debug(
        "Config has attrs: %s",
        {
            k: str(v)
            for k, v in Config.__dict__.items()
            if not isinstance(v, classmethod) and not k.startswith("_")
        },
    )
    logger.debug("Generated get_routes: %s", get_routes)
    logger.debug("Generated post_routes: %s", post_routes)

    app = aiohttp.web.Application()
    app.middlewares.append(
        _create_error_middleware(
            {
                404: _not_found_handler,
                405: _not_allowed_handler,
            }
        )
    )
    app.router.add_route("GET", "/", _root_handler)
    for route in get_routes:
        app.router.add_route("GET", route, _http_handler)
    for route in post_routes:
        app.router.add_route("POST", route, _http_handler)
        app.router.add_route("OPTIONS", route, _http_handler)
    logger.info("Welcome to Nawah")

    try:
        asyncio.run(_web_loop(app))
    except KeyboardInterrupt:
        if time.localtime().tm_hour >= 21 or time.localtime().tm_hour <= 4:
            msg = "night"
        elif time.localtime().tm_hour >= 18:
            msg = "evening"
        elif time.localtime().tm_hour >= 12:
            msg = "afternoon"
        elif time.localtime().tm_hour >= 5:
            msg = "morning"

        logger.info("Have a great %s!", msg)

        sys.exit()



import copy
import logging
from typing import TYPE_CHECKING, Any, MutableMapping, Optional, Tuple, Union

from nawah.classes import Var
from nawah.config import Config

if TYPE_CHECKING:
    from nawah.classes import Attr, Query
    from nawah.types import (NawahQueryOperAnd, NawahQueryOperOr,
                             NawahQuerySpecialGroup, NawahQueryStep)

logger = logging.getLogger("nawah")


def _compile_query(
    *,
    attrs: MutableMapping[str, "Attr"],
    query: "Query",
) -> Tuple[
    Optional[int],
    Optional[int],
    MutableMapping[str, int],
    Optional[list["NawahQuerySpecialGroup"]],
    list[Any],
]:
    aggregate_prefix: list[Any] = []
    aggregate_suffix: list[Any] = []
    aggregate_query: list[Any] = []
    skip: Optional[int] = None
    limit: Optional[int] = None
    sort: MutableMapping[str, int] = {"_id": -1}
    group: Optional[list["NawahQuerySpecialGroup"]] = None
    logger.debug("attempting to process query: %s", query)

    query = copy.deepcopy(query)

    # Update variables per Query Special Args
    if "$deleted" in query and query["$deleted"] is True:
        aggregate_prefix.append({"$match": {"__deleted": {"$exists": True}}})
    else:
        aggregate_prefix.append({"$match": {"__deleted": {"$exists": False}}})
    if "$skip" in query:
        skip = query["$skip"]
    if "$limit" in query:
        limit = query["$limit"]
    if "$sort" in query:
        sort = query["$sort"]
    if "$group" in query:
        group = query["$group"]
    if "$search" in query:
        aggregate_prefix.insert(0, {"$match": {"$text": {"$search": query["$search"]}}})
        project_query: MutableMapping[str, Any] = {
            attr: "$" + attr for attr in attrs.keys()
        }
        project_query["_id"] = "$_id"
        project_query["__score"] = {"$meta": "textScore"}
        aggregate_suffix.append({"$project": project_query})
        aggregate_suffix.append({"$match": {"__score": {"$gt": 0.5}}})
    if "$geo_near" in query:
        aggregate_prefix.insert(
            0,
            {
                "$geoNear": {
                    "near": {
                        "type": "Point",
                        "coordinates": query["$geo_near"]["val"],
                    },
                    "distanceField": query["$geo_near"]["attr"] + ".__distance",
                    "maxDistance": query["$geo_near"]["dist"],
                    "spherical": True,
                }
            },
        )

    for step in query.pipe or []:
        _compile_query_step(
            query=query,
            aggregate_prefix=aggregate_prefix,
            aggregate_suffix=aggregate_suffix,
            aggregate_query=aggregate_query,
            attrs=attrs,
            step=step,
        )

    if "$attrs" in query and isinstance(query["$attrs"], list):
        group_query: dict[str, Any] = {
            "_id": "$_id",
            **{
                attr: {"$first": f"${attr}"}
                for attr in query["$attrs"]
                if attr in attrs.keys()
            },
        }
        aggregate_suffix.append({"$group": group_query})
    else:
        group_query = {
            "_id": "$_id",
            **{attr: {"$first": f"${attr}"} for attr in attrs.keys()},
        }
        aggregate_suffix.append({"$group": group_query})

    logger.debug(
        "processed query, aggregate_prefix:%s, aggregate_suffix:%s, aggregate_match:%s",
        aggregate_prefix,
        aggregate_suffix,
        aggregate_query,
    )

    aggregate_query = aggregate_prefix + aggregate_query + aggregate_suffix
    return (skip, limit, sort, group, aggregate_query)


def _compile_query_step(
    *,
    query: "Query",
    aggregate_prefix: list[Any],
    aggregate_suffix: list[Any],
    aggregate_query: list[Any],
    attrs: MutableMapping[str, "Attr"],
    step: Union["NawahQueryStep", "NawahQueryOperOr", "NawahQueryOperAnd"],
    append_query: bool = True,
) -> None:
    if append_query:
        aggregate_query.append({"$match": step})

    attr_name = list(step)[0]
    attr_val = step[attr_name]  # type: ignore

    if isinstance(attr_val, list):
        for child_attr_val in attr_val:
            _compile_query_step(
                query=query,
                aggregate_prefix=aggregate_prefix,
                aggregate_suffix=aggregate_suffix,
                aggregate_query=aggregate_query,
                attrs=attrs,
                step=child_attr_val,
                append_query=False,
            )
        return

    if "." not in attr_name:
        return

    root_attr_name = attr_name.split(".")[0]

    if root_attr_name not in attrs or not attrs[root_attr_name].extn:
        return

    # [TODO] Check if this works with EXTN as Attr Type TYPE
    # Don't attempt to extn attr that is already extended
    lookup_query = False
    for stage in aggregate_prefix:
        if "$lookup" in stage.keys() and stage["$lookup"]["as"] == root_attr_name:
            lookup_query = True
            break
    if not lookup_query:
        if isinstance(attrs[root_attr_name].extn.module, Var):
            extn_collection = (
                f'{query[f"{attrs[root_attr_name].extn.module.var}:$eq"][0]}_docs'
            )
        else:
            extn_collection = Config.modules[
                attrs[root_attr_name].extn.module
            ].collection
        aggregate_prefix.append(
            {"$addFields": {root_attr_name: {"$toObjectId": f"${root_attr_name}"}}}
        )
        aggregate_prefix.append(
            {
                "$lookup": {
                    "from": extn_collection,
                    "localField": root_attr_name,
                    "foreignField": "_id",
                    "as": root_attr_name,
                }
            }
        )
        aggregate_prefix.append({"$unwind": f"${root_attr_name}"})
        group_query: MutableMapping[str, Any] = {
            attr: {"$first": f"${attr}"} for attr in attrs.keys()
        }
        group_query[root_attr_name] = {"$first": f"${root_attr_name}._id"}
        group_query["_id"] = "$_id"
        aggregate_suffix.append({"$group": group_query})
        aggregate_suffix.append(
            {"$addFields": {root_attr_name: {"$toString": f"${root_attr_name}"}}}
        )



"""Provides utilties used to manage Nawah App serving"""

import logging

from nawah.config import Config

logger = logging.getLogger("nawah")


def _populate_routes():
    get_routes = []
    post_routes = []
    for module_name, module in Config.modules.items():
        for func_name, func in module.funcs.items():
            if func.get_func:
                for get_args_set in func.query_attrs or [{}]:
                    get_args = ""
                    if get_args_set:
                        for query_attr in get_args_set:
                            # [TODO] Deprecated check. Remove with release
                            if ":" not in query_attr:
                                get_args += f"/{{{query_attr}}}"
                            else:
                                if query_attr.split(":")[1] != "$eq":
                                    raise Exception(
                                        "Nawah Function '{module_name}.{func_name}' set exposed as "
                                        "GET endpoint must only define Query Attrs with Query "
                                        "Operators $eq"
                                    )
                                get_args += f"/{{{query_attr.split(':')[0]}}}"

                    get_routes.append(f"/{module_name}/{func_name}{get_args}")

            func.post_func = True
            if func.post_func:
                post_routes.append(f"/{module_name}/{func_name}")

    return (get_routes, post_routes)



"""Provides IOC-style classes for 'create_file' Base Function callable, 'set_file' Utility"""

import datetime
import logging
from typing import TYPE_CHECKING, MutableMapping, Protocol

from nawah.classes import Attr
from nawah.enums import AttrType
from nawah.exceptions import InvalidAttrTypeException
from nawah.utils import extract_attr, validate_doc

from .._shared import FILE_ATTRS
from ..exceptions import NoDocUpdatedException, UtilityModuleDataCallException

if TYPE_CHECKING:
    from nawah.classes import Module
    from nawah.types import (DataCreateCallable, DataUpdateCallable, NawahDoc,
                             NawahSession, Results)

logger = logging.getLogger("nawah")


class SetFileCallable(Protocol):
    """Provides type-hint for 'set_file_callable' of 'CreateFile'"""

    # pylint: disable=too-few-public-methods

    async def __call__(
        self,
        *,
        module_name: str,
        session: "NawahSession",
        doc: "NawahDoc",
        raise_no_success: bool,
    ) -> "Results":
        ...


class CreateFile:
    """IOC-style implementation for 'create_file' Base Function"""

    # pylint: disable=too-few-public-methods

    __modules: MutableMapping[str, "Module"]
    __data_create_callable: "DataCreateCallable"
    __set_file_callable: "SetFileCallable"

    def __init__(
        self,
        modules: MutableMapping[str, "Module"],
        data_create_callable: "DataCreateCallable",
        set_file_callable: "SetFileCallable",
    ):
        self.__modules = modules
        self.__data_create_callable = data_create_callable
        self.__set_file_callable = set_file_callable

    async def __call__(
        self,
        *,
        module_name: str,
        session: "NawahSession",
        doc: "NawahDoc",
    ) -> "Results":
        module = self.__modules[module_name]

        if not module.collection:
            raise UtilityModuleDataCallException(
                module_name=module_name, func_name="create_file"
            )

        if "_id" in doc:
            return await self.__set_file_callable(
                module_name=module_name,
                session=session,
                doc=doc,
                raise_no_success=True,
            )

        file_attr = extract_attr(attrs=module.attrs, path=doc["attr"])

        # For list of files, assert call points to LIST attr, and it contains a FILE attr
        if file_attr.type == AttrType.LIST:
            if file_attr.args["list"][0].type != AttrType.FILE:
                raise InvalidAttrTypeException(
                    attr_type=file_attr.type,
                )

            file_attr = file_attr.args["list"][0]

        # If it is still not a FILE attr raise InvalidAttrTypeException
        if file_attr.type != AttrType.FILE:
            raise InvalidAttrTypeException(attr_type=file_attr.type)

        types_pattern = None
        if file_attr.args["types"]:
            types_pattern = "|".join(file_attr.args["types"])
            # Replace astrisks (*) with dot-astrisks to mark it as valid RegExp pattern
            types_pattern = types_pattern.replace("*", ".*")

        file_doc_attrs = {
            **FILE_ATTRS,
            "file": Attr.TYPED_DICT(
                dict={
                    "name": Attr.STR(),
                    "lastModified": Attr.INT(),
                    "type": Attr.STR(pattern=types_pattern),
                    "size": Attr.INT(),
                    "content": Attr(
                        desc="__sys_attr",
                        type=AttrType.BYTES,
                        args={},
                    ),
                }
            ),
        }

        file_doc = {
            "user": session["user"]["_id"],
            "doc": "000000000000000000000000",
            "attr": doc["attr"],
            "file": {
                "name": doc["name"],
                "lastModified": doc["lastModified"],
                "type": doc["type"],
                "size": len(doc["content"]),
                "content": doc["content"],
            },
            "create_time": datetime.datetime.utcnow().isoformat(),
        }

        validate_doc(mode="create", attrs=file_doc_attrs, doc=file_doc)

        # Execute Data driver create
        results = await self.__data_create_callable(
            session=session, collection_name=f"{module.collection}__file", doc=file_doc
        )

        return {
            "status": 200,
            "msg": f'Created {results["count"]} files',
            "args": results,
        }


class SetFile:
    """IOC-style implementation for 'set_file' Utility"""

    # pylint: disable=too-few-public-methods

    __modules: MutableMapping[str, "Module"]
    __data_update_callable: "DataUpdateCallable"

    def __init__(
        self,
        modules: MutableMapping[str, "Module"],
        data_update_callable: "DataUpdateCallable",
    ):
        self.__modules = modules
        self.__data_update_callable = data_update_callable

    async def __call__(
        self,
        *,
        module_name: str,
        session: "NawahSession",
        doc: "NawahDoc",
        raise_no_success: bool,
    ) -> "Results":
        module = self.__modules[module_name]

        update_results = await self.__data_update_callable(
            session=session,
            collection_name=f"{module.collection}__file",
            docs=[doc["_id"]],
            doc={"doc": doc["doc"]},
        )

        if raise_no_success is True and update_results["count"] == 0:
            raise NoDocUpdatedException(module_name=module_name)

        return {
            "status": 200,
            "msg": f"Updated {update_results['count']} docs",
            "args": update_results,
        }



"""Provides types used in Nawah data module"""

from typing import TYPE_CHECKING, MutableSequence, Protocol, Union

if TYPE_CHECKING:
    from bson import ObjectId

    from ._types import NawahDoc, NawahSession, ResultsArgs


class DataCreateCallable(Protocol):
    """Provides type-hint for implementation of Data.create to use with IOC classes"""

    # pylint: disable=too-few-public-methods

    async def __call__(
        self,
        *,
        session: "NawahSession",
        collection_name: str,
        doc: "NawahDoc",
    ) -> "ResultsArgs":
        ...


class DataUpdateCallable(Protocol):
    """Provides type-hint for implementation of Data.update to use with IOC classes"""

    # pylint: disable=too-few-public-methods

    async def __call__(
        self,
        *,
        session: "NawahSession",
        collection_name: str,
        docs: MutableSequence[Union[str, "ObjectId"]],
        doc: "NawahDoc",
    ) -> "ResultsArgs":
        ...
